@@ -27,6 +27,7 @@
 #include "print-tree.h"
 #include "transaction.h"
 #include "list.h"
+#include "version.h"
 
 static u64 bytes_used = 0;
 static u64 total_csum_bytes = 0;
@@ -119,15 +120,23 @@ static int check_leaf(struct btrfs_root *root,
 	for (i = 0; nritems > 1 && i < nritems - 2; i++) {
 		btrfs_item_key(buf, &key, i);
 		btrfs_item_key_to_cpu(buf, &cpukey, i + 1);
-		if (btrfs_comp_keys(&key, &cpukey) >= 0)
+		if (btrfs_comp_keys(&key, &cpukey) >= 0) {
+			fprintf(stderr, "bad key ordering %d %d\n", i, i+1);
 			return 1;
+		}
 		if (btrfs_item_offset_nr(buf, i) !=
-			btrfs_item_end_nr(buf, i + 1))
+			btrfs_item_end_nr(buf, i + 1)) {
+			fprintf(stderr, "incorrect offsets %u %u\n",
+				btrfs_item_offset_nr(buf, i),
+				btrfs_item_end_nr(buf, i + 1));
+			return 1;
+		}
+		if (i == 0 && btrfs_item_end_nr(buf, i) !=
+		    BTRFS_LEAF_DATA_SIZE(root)) {
+			fprintf(stderr, "bad item end %u wanted %u\n",
+				btrfs_item_end_nr(buf, i),
+				(unsigned)BTRFS_LEAF_DATA_SIZE(root));
 			return 1;
-		if (i == 0) {
-			if (btrfs_item_end_nr(buf, i) !=
-			       BTRFS_LEAF_DATA_SIZE(root))
-				return 1;
 		}
 	}
 	return 0;
@@ -472,7 +481,6 @@ static int pick_next_pending(struct cache_tree *pending,
 	}
 	return ret;
 }
-static struct extent_buffer reada_buf;
 
 static int run_next_block(struct btrfs_root *root,
 			  struct block_info *bits,
@@ -495,7 +503,6 @@ static int run_next_block(struct btrfs_root *root,
 	struct cache_extent *cache;
 	int reada_bits;
 
-	u64 last_block = 0;
 	ret = pick_next_pending(pending, reada, nodes, *last, bits,
 				bits_nr, &reada_bits);
 	if (ret == 0) {
@@ -503,14 +510,12 @@ static int run_next_block(struct btrfs_root *root,
 	}
 	if (!reada_bits) {
 		for(i = 0; i < ret; i++) {
-			u64 offset;
 			insert_cache_extent(reada, bits[i].start,
 					    bits[i].size);
-			btrfs_map_bh_to_logical(root, &reada_buf,
-						bits[i].start);
-			offset = reada_buf.dev_bytenr;
-			last_block = bits[i].start;
-			readahead(reada_buf.fd, offset, bits[i].size);
+
+			/* fixme, get the parent transid */
+			readahead_tree_block(root, bits[i].start,
+					     bits[i].size, 0);
 		}
 	}
 	*last = bits[0].start;
@@ -533,7 +538,8 @@ static int run_next_block(struct btrfs_root *root,
 		free(cache);
 	}
 
-	buf = read_tree_block(root, bytenr, size);
+	/* fixme, get the real parent transid */
+	buf = read_tree_block(root, bytenr, size, 0);
 	nritems = btrfs_header_nritems(buf);
 	ret = check_block(root, extent_cache, buf);
 	if (ret) {
@@ -605,6 +611,9 @@ static int run_next_block(struct btrfs_root *root,
 
 			data_bytes_allocated +=
 				btrfs_file_extent_disk_num_bytes(buf, fi);
+			if (data_bytes_allocated < root->sectorsize) {
+				abort();
+			}
 			data_bytes_referenced +=
 				btrfs_file_extent_num_bytes(buf, fi);
 			ret = add_extent_rec(extent_cache, NULL, bytenr,
@@ -711,6 +720,7 @@ int check_extent_refs(struct btrfs_root *root,
 
 void print_usage(void) {
 	fprintf(stderr, "usage: btrfsck dev\n");
+	fprintf(stderr, "%s\n", BTRFS_BUILD_VERSION);
 	exit(1);
 }
 
@@ -742,7 +752,7 @@ int main(int ac, char **av) {
 	cache_tree_init(&nodes);
 	cache_tree_init(&reada);
 
-	root = open_ctree(av[1], 0);
+	root = open_ctree(av[1], 0, 0);
 
 	bits_nr = 1024;
 	bits = malloc(bits_nr * sizeof(struct block_info));
@@ -755,6 +765,10 @@ int main(int ac, char **av) {
 			    &extent_cache, &pending, &seen, &reada, &nodes,
 			    root->fs_info->tree_root->root_key.objectid);
 
+	add_root_to_pending(root->fs_info->chunk_root->node, bits, bits_nr,
+			    &extent_cache, &pending, &seen, &reada, &nodes,
+			    root->fs_info->chunk_root->root_key.objectid);
+
 	btrfs_init_path(&path);
 	key.offset = 0;
 	key.objectid = 0;
@@ -782,7 +796,7 @@ int main(int ac, char **av) {
 			buf = read_tree_block(root->fs_info->tree_root,
 					      btrfs_root_bytenr(&ri),
 					      btrfs_level_size(root,
-						       btrfs_root_level(&ri)));
+					       btrfs_root_level(&ri)), 0);
 			add_root_to_pending(buf, bits, bits_nr, &extent_cache,
 					    &pending, &seen, &reada, &nodes,
 					    found_key.objectid);
@@ -809,5 +823,6 @@ int main(int ac, char **av) {
 	printf("file data blocks allocated: %llu\n referenced %llu\n",
 		(unsigned long long)data_bytes_allocated,
 		(unsigned long long)data_bytes_referenced);
+	printf("%s\n", BTRFS_BUILD_VERSION);
 	return ret;
 }
