@@ -32,6 +32,7 @@
 #include "ctree.h"
 #include "transaction.h"
 #include "utils.h"
+#include "version.h"
 
 #ifdef __CHECKER__
 #define BLKGETSIZE64 0
@@ -43,14 +44,17 @@ static inline int ioctl(int fd, int define, void *arg) { return 0; }
 
 void print_usage(void)
 {
-	printf("usage: btrfsctl [ -s name ] [-d] [-r size] file_or_dir\n");
-	printf("\t-d filename defragments one file\n");
-	printf("\t-d directory defragments the entire Btree\n");
-	printf("\t-s snap_name existing_subvol creates a new snapshot\n");
-	printf("\t-s snap_name tree_root creates a new subvolume\n");
-	printf("\t-r [+-]size[gkm] resize the FS\n");
-	printf("\t-A device scans the device for a Btrfs filesystem\n");
-	printf("\t-a scans all devices for a Btrfs filesystems\n");
+	printf("usage: btrfsctl [ -d file|dir] [ -s snap_name subvol|tree ]\n");
+	printf("                [-r size] [-A device] [-a] [-c]\n");
+	printf("\t-d filename: defragments one file\n");
+	printf("\t-d directory: defragments the entire Btree\n");
+	printf("\t-s snap_name: existing_subvol creates a new snapshot\n");
+	printf("\t-s snap_name: tree_root creates a new subvolume\n");
+	printf("\t-r [+-]size[gkm]: resize the FS by size amount\n");
+	printf("\t-A device: scans the device file for a Btrfs filesystem\n");
+	printf("\t-a: scans all devices for Btrfs filesystems\n");
+	printf("\t-c: forces a single FS sync\n");
+	printf("%s\n", BTRFS_BUILD_VERSION);
 	exit(1);
 }
 
@@ -72,7 +76,7 @@ int main(int ac, char **av)
 		btrfs_scan_one_dir("/dev", 1);
 		exit(0);
 	}
-	for (i = 1; i < ac - 1; i++) {
+	for (i = 1; i < ac; i++) {
 		if (strcmp(av[i], "-s") == 0) {
 			if (i + 1 >= ac - 1) {
 				fprintf(stderr, "-s requires an arg");
@@ -115,10 +119,13 @@ int main(int ac, char **av)
 				exit(1);
 			}
 			command = BTRFS_IOC_RESIZE;
+		} else if (strcmp(av[i], "-c") == 0) {
+			command = BTRFS_IOC_SYNC;
 		}
 	}
 	if (command == 0) {
 		fprintf(stderr, "no valid commands given\n");
+		print_usage();
 		exit(1);
 	}
 	fname = av[ac - 1];
@@ -149,7 +156,19 @@ int main(int ac, char **av)
 	else
 		args.name[0] = '\0';
 	ret = ioctl(fd, command, &args);
-	printf("ioctl returns %d\n", ret);
-	return 0;
+	if (ret < 0) {
+		perror("ioctl:");
+		exit(1);
+	}
+	if (ret == 0) {
+		printf("operation complete\n");
+	} else {
+		printf("ioctl failed with error %d\n", ret);
+	}
+	printf("%s\n", BTRFS_BUILD_VERSION);
+	if (ret)
+		exit(0);
+	else
+		exit(1);
 }
 
