@@ -18,6 +18,7 @@
 
 #define _XOPEN_SOURCE 600
 #define __USE_XOPEN2K
+#define _GNU_SOURCE 1
 #include <stdio.h>
 #include <stdlib.h>
 #include <sys/types.h>
@@ -28,26 +29,21 @@
 #include "radix-tree.h"
 #include "ctree.h"
 #include "disk-io.h"
+#include "volumes.h"
 #include "transaction.h"
 #include "crc32c.h"
-
-int btrfs_map_bh_to_logical(struct btrfs_root *root, struct extent_buffer *buf,
-			    u64 logical)
-{
-	buf->fd = root->fs_info->fp;
-	buf->dev_bytenr = logical;
-	return 0;
-}
+#include "utils.h"
+#include "print-tree.h"
 
 static int check_tree_block(struct btrfs_root *root, struct extent_buffer *buf)
 {
 	if (buf->start != btrfs_header_bytenr(buf))
-		BUG();
+		return 1;
 
 	if (memcmp_extent_buffer(buf, root->fs_info->fsid,
 				 (unsigned long)btrfs_header_fsid(buf),
 				 BTRFS_FSID_SIZE))
-		BUG();
+		return 1;
 	return 0;
 }
 
@@ -61,8 +57,8 @@ void btrfs_csum_final(u32 crc, char *result)
 	*(__le32 *)result = ~cpu_to_le32(crc);
 }
 
-static int csum_tree_block(struct btrfs_root *root, struct extent_buffer *buf,
-			   int verify)
+int csum_tree_block(struct btrfs_root *root, struct extent_buffer *buf,
+		    int verify)
 {
 	char result[BTRFS_CRC32_SIZE];
 	u32 len;
@@ -74,7 +70,9 @@ static int csum_tree_block(struct btrfs_root *root, struct extent_buffer *buf,
 
 	if (verify) {
 		if (memcmp_extent_buffer(buf, result, 0, BTRFS_CRC32_SIZE)) {
-			printk("checksum verify failed on %llu\n", buf->start);
+			printk("checksum verify failed on %llu wanted %X "
+			       "found %X\n", (unsigned long long)buf->start,
+			       *((int *)result), *((int *)buf));
 			return 1;
 		}
 	} else {
@@ -97,42 +95,143 @@ struct extent_buffer *btrfs_find_create_tree_block(struct btrfs_root *root,
 				   blocksize);
 }
 
-int readahead_tree_block(struct btrfs_root *root, u64 bytenr, u32 blocksize)
+int readahead_tree_block(struct btrfs_root *root, u64 bytenr, u32 blocksize,
+			 u64 parent_transid)
 {
+	int ret;
+	int dev_nr;
+	struct extent_buffer *eb;
+	u64 length;
+	struct btrfs_multi_bio *multi = NULL;
+	struct btrfs_device *device;
+
+	eb = btrfs_find_tree_block(root, bytenr, blocksize);
+	if (eb && btrfs_buffer_uptodate(eb, parent_transid)) {
+		free_extent_buffer(eb);
+		return 0;
+	}
+
+	dev_nr = 0;
+	length = blocksize;
+	ret = btrfs_map_block(&root->fs_info->mapping_tree, READ,
+			      bytenr, &length, &multi, 0);
+	BUG_ON(ret);
+	device = multi->stripes[0].dev;
+	device->total_ios++;
+	blocksize = min(blocksize, (u32)(64 * 1024));
+	readahead(device->fd, multi->stripes[0].physical, blocksize);
+	kfree(multi);
 	return 0;
 }
 
-struct extent_buffer *read_tree_block(struct btrfs_root *root, u64 bytenr,
-				     u32 blocksize)
+static int verify_parent_transid(struct extent_io_tree *io_tree,
+				 struct extent_buffer *eb, u64 parent_transid)
 {
 	int ret;
+
+	if (!parent_transid || btrfs_header_generation(eb) == parent_transid)
+		return 0;
+
+	if (extent_buffer_uptodate(eb) &&
+	    btrfs_header_generation(eb) == parent_transid) {
+		ret = 0;
+		goto out;
+	}
+	printk("parent transid verify failed on %llu wanted %llu found %llu\n",
+	       (unsigned long long)eb->start,
+	       (unsigned long long)parent_transid,
+	       (unsigned long long)btrfs_header_generation(eb));
+	ret = 1;
+out:
+	clear_extent_buffer_uptodate(io_tree, eb);
+	return ret;
+
+}
+
+
+struct extent_buffer *read_tree_block(struct btrfs_root *root, u64 bytenr,
+				     u32 blocksize, u64 parent_transid)
+{
+	int ret;
+	int dev_nr;
 	struct extent_buffer *eb;
+	u64 length;
+	struct btrfs_multi_bio *multi = NULL;
+	struct btrfs_device *device;
+	int mirror_num = 0;
+	int num_copies;
 
 	eb = btrfs_find_create_tree_block(root, bytenr, blocksize);
 	if (!eb)
 		return NULL;
-	if (!btrfs_buffer_uptodate(eb)) {
-		btrfs_map_bh_to_logical(root, eb, eb->start);
+
+	if (btrfs_buffer_uptodate(eb, parent_transid))
+		return eb;
+
+	dev_nr = 0;
+	length = blocksize;
+	while (1) {
+		ret = btrfs_map_block(&root->fs_info->mapping_tree, READ,
+				      eb->start, &length, &multi, mirror_num);
+		BUG_ON(ret);
+		device = multi->stripes[0].dev;
+		eb->fd = device->fd;
+		device->total_ios++;
+		eb->dev_bytenr = multi->stripes[0].physical;
+		kfree(multi);
 		ret = read_extent_from_disk(eb);
-		if (ret) {
-			free_extent_buffer(eb);
-			return NULL;
+		if (ret == 0 && check_tree_block(root, eb) == 0 &&
+		    csum_tree_block(root, eb, 1) == 0 &&
+		    verify_parent_transid(eb->tree, eb, parent_transid) == 0) {
+			btrfs_set_buffer_uptodate(eb);
+			return eb;
+		}
+		num_copies = btrfs_num_copies(&root->fs_info->mapping_tree,
+					      eb->start, eb->len);
+		if (num_copies == 1) {
+			break;
+		}
+		mirror_num++;
+		if (mirror_num > num_copies) {
+			break;
 		}
-		btrfs_set_buffer_uptodate(eb);
 	}
-	return eb;
+	free_extent_buffer(eb);
+	return NULL;
 }
 
 int write_tree_block(struct btrfs_trans_handle *trans, struct btrfs_root *root,
 		     struct extent_buffer *eb)
 {
+	int ret;
+	int dev_nr;
+	u64 length;
+	struct btrfs_multi_bio *multi = NULL;
+
 	if (check_tree_block(root, eb))
 		BUG();
-	if (!btrfs_buffer_uptodate(eb))
+	if (!btrfs_buffer_uptodate(eb, trans->transid))
 		BUG();
-	btrfs_map_bh_to_logical(root, eb, eb->start);
+
+	btrfs_set_header_flag(eb, BTRFS_HEADER_FLAG_WRITTEN);
 	csum_tree_block(root, eb, 0);
-	return write_extent_to_disk(eb);
+
+	dev_nr = 0;
+	length = eb->len;
+	ret = btrfs_map_block(&root->fs_info->mapping_tree, WRITE,
+			      eb->start, &length, &multi, 0);
+
+	while(dev_nr < multi->num_stripes) {
+		BUG_ON(ret);
+		eb->fd = multi->stripes[dev_nr].dev->fd;
+		eb->dev_bytenr = multi->stripes[dev_nr].physical;
+		multi->stripes[dev_nr].dev->total_ios++;
+		dev_nr++;
+		ret = write_extent_to_disk(eb);
+		BUG_ON(ret);
+	}
+	kfree(multi);
+	return 0;
 }
 
 static int __setup_root(u32 nodesize, u32 leafsize, u32 sectorsize,
@@ -146,39 +245,56 @@ static int __setup_root(u32 nodesize, u32 leafsize, u32 sectorsize,
 	root->leafsize = leafsize;
 	root->stripesize = stripesize;
 	root->ref_cows = 0;
+	root->track_dirty = 0;
+
 	root->fs_info = fs_info;
 	root->objectid = objectid;
 	root->last_trans = 0;
 	root->highest_inode = 0;
 	root->last_inode_alloc = 0;
+
+	INIT_LIST_HEAD(&root->dirty_list);
 	memset(&root->root_key, 0, sizeof(root->root_key));
 	memset(&root->root_item, 0, sizeof(root->root_item));
 	root->root_key.objectid = objectid;
 	return 0;
 }
 
+static int update_cowonly_root(struct btrfs_trans_handle *trans,
+			       struct btrfs_root *root)
+{
+	int ret;
+	u64 old_root_bytenr;
+	struct btrfs_root *tree_root = root->fs_info->tree_root;
+
+	btrfs_write_dirty_block_groups(trans, root);
+	while(1) {
+		old_root_bytenr = btrfs_root_bytenr(&root->root_item);
+		if (old_root_bytenr == root->node->start)
+			break;
+		btrfs_set_root_bytenr(&root->root_item,
+				       root->node->start);
+		root->root_item.level = btrfs_header_level(root->node);
+		ret = btrfs_update_root(trans, tree_root,
+					&root->root_key,
+					&root->root_item);
+		BUG_ON(ret);
+		btrfs_write_dirty_block_groups(trans, root);
+	}
+	return 0;
+}
+
 static int commit_tree_roots(struct btrfs_trans_handle *trans,
 			     struct btrfs_fs_info *fs_info)
 {
-	int ret;
-	u64 old_extent_bytenr;
-	struct btrfs_root *tree_root = fs_info->tree_root;
-	struct btrfs_root *extent_root = fs_info->extent_root;
+	struct btrfs_root *root;
+	struct list_head *next;
 
-	btrfs_write_dirty_block_groups(trans, fs_info->extent_root);
-	while(1) {
-		old_extent_bytenr = btrfs_root_bytenr(&extent_root->root_item);
-		if (old_extent_bytenr == extent_root->node->start)
-			break;
-		btrfs_set_root_bytenr(&extent_root->root_item,
-				       extent_root->node->start);
-		extent_root->root_item.level =
-			btrfs_header_level(extent_root->node);
-		ret = btrfs_update_root(trans, tree_root,
-					&extent_root->root_key,
-					&extent_root->root_item);
-		BUG_ON(ret);
-		btrfs_write_dirty_block_groups(trans, fs_info->extent_root);
+	while(!list_empty(&fs_info->dirty_cowonly_roots)) {
+		next = fs_info->dirty_cowonly_roots.next;
+		list_del_init(next);
+		root = list_entry(next, struct btrfs_root, dirty_list);
+		update_cowonly_root(trans, root);
 	}
 	return 0;
 }
@@ -189,7 +305,7 @@ static int __commit_transaction(struct btrfs_trans_handle *trans,
 	u64 start;
 	u64 end;
 	struct extent_buffer *eb;
-	struct extent_map_tree *tree = &root->fs_info->extent_cache;
+	struct extent_io_tree *tree = &root->fs_info->extent_cache;
 	int ret;
 
 	while(1) {
@@ -289,7 +405,7 @@ static int find_and_setup_root(struct btrfs_root *tree_root,
 
 	blocksize = btrfs_level_size(root, btrfs_root_level(&root->root_item));
 	root->node = read_tree_block(root, btrfs_root_bytenr(&root->root_item),
-				     blocksize);
+				     blocksize, 0);
 	BUG_ON(!root->node);
 	return 0;
 }
@@ -356,25 +472,34 @@ out:
 	}
 	blocksize = btrfs_level_size(root, btrfs_root_level(&root->root_item));
 	root->node = read_tree_block(root, btrfs_root_bytenr(&root->root_item),
-				     blocksize);
+				     blocksize, 0);
 	BUG_ON(!root->node);
 insert:
 	root->ref_cows = 1;
 	return root;
 }
 
-struct btrfs_root *open_ctree(char *filename, u64 sb_bytenr)
+struct btrfs_root *open_ctree(const char *filename, u64 sb_bytenr, int writes)
 {
 	int fp;
+	struct btrfs_root *root;
+	int flags = O_CREAT | O_RDWR;
 
-	fp = open(filename, O_CREAT | O_RDWR, 0600);
+	if (!writes)
+		flags = O_RDONLY;
+
+	fp = open(filename, flags, 0600);
 	if (fp < 0) {
 		return NULL;
 	}
-	return open_ctree_fd(fp, sb_bytenr);
+	root = open_ctree_fd(fp, filename, sb_bytenr, writes);
+	close(fp);
+
+	return root;
 }
 
-struct btrfs_root *open_ctree_fd(int fp, u64 sb_bytenr)
+struct btrfs_root *open_ctree_fd(int fp, const char *path, u64 sb_bytenr,
+				 int writes)
 {
 	u32 sectorsize;
 	u32 nodesize;
@@ -384,34 +509,74 @@ struct btrfs_root *open_ctree_fd(int fp, u64 sb_bytenr)
 	struct btrfs_root *root = malloc(sizeof(struct btrfs_root));
 	struct btrfs_root *tree_root = malloc(sizeof(struct btrfs_root));
 	struct btrfs_root *extent_root = malloc(sizeof(struct btrfs_root));
+	struct btrfs_root *chunk_root = malloc(sizeof(struct btrfs_root));
+	struct btrfs_root *dev_root = malloc(sizeof(struct btrfs_root));
 	struct btrfs_fs_info *fs_info = malloc(sizeof(*fs_info));
 	int ret;
 	struct btrfs_super_block *disk_super;
+	struct btrfs_fs_devices *fs_devices = NULL;
+	u64 total_devs;
 
 	if (sb_bytenr == 0)
 		sb_bytenr = BTRFS_SUPER_INFO_OFFSET;
 
-	fs_info->fp = fp;
-	fs_info->running_transaction = NULL;
+	ret = btrfs_scan_one_device(fp, path, &fs_devices,
+				    &total_devs, sb_bytenr);
+
+	if (ret) {
+		fprintf(stderr, "No valid Btrfs found on %s\n", path);
+		return NULL;
+	}
+
+	if (total_devs != 1) {
+		ret = btrfs_scan_for_fsid(fs_devices, total_devs, 1);
+		BUG_ON(ret);
+	}
+
+	memset(fs_info, 0, sizeof(*fs_info));
 	fs_info->fs_root = root;
 	fs_info->tree_root = tree_root;
 	fs_info->extent_root = extent_root;
-	fs_info->extent_ops = NULL;
-	fs_info->priv_data = NULL;
-	extent_map_tree_init(&fs_info->extent_cache);
-	extent_map_tree_init(&fs_info->free_space_cache);
-	extent_map_tree_init(&fs_info->pending_tree);
-	extent_map_tree_init(&fs_info->pinned_extents);
-	extent_map_tree_init(&fs_info->del_pending);
-	extent_map_tree_init(&fs_info->block_group_cache);
+	fs_info->chunk_root = chunk_root;
+	fs_info->dev_root = dev_root;
+
+	if (!writes)
+		fs_info->readonly = 1;
+
+	extent_io_tree_init(&fs_info->extent_cache);
+	extent_io_tree_init(&fs_info->free_space_cache);
+	extent_io_tree_init(&fs_info->block_group_cache);
+	extent_io_tree_init(&fs_info->pinned_extents);
+	extent_io_tree_init(&fs_info->pending_del);
+	extent_io_tree_init(&fs_info->extent_ins);
+
+	cache_tree_init(&fs_info->mapping_tree.cache_tree);
 
 	mutex_init(&fs_info->fs_mutex);
+	fs_info->fs_devices = fs_devices;
+	INIT_LIST_HEAD(&fs_info->dirty_cowonly_roots);
+	INIT_LIST_HEAD(&fs_info->space_info);
 
-	__setup_root(512, 512, 512, 512, tree_root,
+	__setup_root(4096, 4096, 4096, 4096, tree_root,
 		     fs_info, BTRFS_ROOT_TREE_OBJECTID);
 
-	fs_info->sb_buffer = read_tree_block(tree_root, sb_bytenr, 512);
+	if (writes)
+		ret = btrfs_open_devices(fs_devices, O_RDWR);
+	else
+		ret = btrfs_open_devices(fs_devices, O_RDONLY);
+	BUG_ON(ret);
+
+	ret = btrfs_bootstrap_super_map(&fs_info->mapping_tree, fs_devices);
+	BUG_ON(ret);
+	fs_info->sb_buffer = btrfs_find_create_tree_block(tree_root, sb_bytenr,
+							  4096);
 	BUG_ON(!fs_info->sb_buffer);
+	fs_info->sb_buffer->fd = fs_devices->latest_bdev;
+	fs_info->sb_buffer->dev_bytenr = sb_bytenr;
+	ret = read_extent_from_disk(fs_info->sb_buffer);
+	BUG_ON(ret);
+	btrfs_set_buffer_uptodate(fs_info->sb_buffer);
+
 	read_extent_buffer(fs_info->sb_buffer, &fs_info->super_copy, 0,
 			   sizeof(fs_info->super_copy));
 	read_extent_buffer(fs_info->sb_buffer, fs_info->fsid,
@@ -433,44 +598,139 @@ struct btrfs_root *open_ctree_fd(int fp, u64 sb_bytenr)
 	tree_root->sectorsize = sectorsize;
 	tree_root->stripesize = stripesize;
 
+	ret = btrfs_read_super_device(tree_root, fs_info->sb_buffer);
+	BUG_ON(ret);
+	ret = btrfs_read_sys_array(tree_root);
+	BUG_ON(ret);
+	blocksize = btrfs_level_size(tree_root,
+				     btrfs_super_chunk_root_level(disk_super));
+
+	__setup_root(nodesize, leafsize, sectorsize, stripesize,
+		     chunk_root, fs_info, BTRFS_CHUNK_TREE_OBJECTID);
+
+	chunk_root->node = read_tree_block(chunk_root,
+					   btrfs_super_chunk_root(disk_super),
+					   blocksize, 0);
+
+	BUG_ON(!chunk_root->node);
+
+	read_extent_buffer(chunk_root->node, fs_info->chunk_tree_uuid,
+	         (unsigned long)btrfs_header_chunk_tree_uuid(chunk_root->node),
+		 BTRFS_UUID_SIZE);
+
+	ret = btrfs_read_chunk_tree(chunk_root);
+	BUG_ON(ret);
+
 	blocksize = btrfs_level_size(tree_root,
 				     btrfs_super_root_level(disk_super));
+
 	tree_root->node = read_tree_block(tree_root,
 					  btrfs_super_root(disk_super),
-					  blocksize);
+					  blocksize, 0);
 	BUG_ON(!tree_root->node);
 	ret = find_and_setup_root(tree_root, fs_info,
 				  BTRFS_EXTENT_TREE_OBJECTID, extent_root);
 	BUG_ON(ret);
+	extent_root->track_dirty = 1;
+
+	ret = find_and_setup_root(tree_root, fs_info,
+				  BTRFS_DEV_TREE_OBJECTID, dev_root);
+	BUG_ON(ret);
+	dev_root->track_dirty = 1;
+
 	ret = find_and_setup_root(tree_root, fs_info,
 				  BTRFS_FS_TREE_OBJECTID, root);
 	BUG_ON(ret);
 	root->ref_cows = 1;
 	fs_info->generation = btrfs_super_generation(disk_super) + 1;
 	btrfs_read_block_groups(root);
+
+	fs_info->data_alloc_profile = (u64)-1;
+	fs_info->metadata_alloc_profile = (u64)-1;
+	fs_info->system_alloc_profile = fs_info->metadata_alloc_profile;
+
 	return root;
 }
 
+int write_all_supers(struct btrfs_root *root)
+{
+	struct list_head *cur;
+	struct list_head *head = &root->fs_info->fs_devices->devices;
+	struct btrfs_device *dev;
+	struct extent_buffer *sb;
+	struct btrfs_dev_item *dev_item;
+	int ret;
+
+	sb = root->fs_info->sb_buffer;
+	dev_item = (struct btrfs_dev_item *)offsetof(struct btrfs_super_block,
+						      dev_item);
+	list_for_each(cur, head) {
+		dev = list_entry(cur, struct btrfs_device, dev_list);
+		btrfs_set_device_type(sb, dev_item, dev->type);
+		btrfs_set_device_id(sb, dev_item, dev->devid);
+		btrfs_set_device_total_bytes(sb, dev_item, dev->total_bytes);
+		btrfs_set_device_bytes_used(sb, dev_item, dev->bytes_used);
+		btrfs_set_device_io_align(sb, dev_item, dev->io_align);
+		btrfs_set_device_io_width(sb, dev_item, dev->io_width);
+		btrfs_set_device_sector_size(sb, dev_item, dev->sector_size);
+		write_extent_buffer(sb, dev->uuid,
+				    (unsigned long)btrfs_device_uuid(dev_item),
+				    BTRFS_UUID_SIZE);
+		sb->fd = dev->fd;
+		sb->dev_bytenr = sb->start;
+		btrfs_set_header_flag(sb, BTRFS_HEADER_FLAG_WRITTEN);
+		csum_tree_block(root, sb, 0);
+		ret = write_extent_to_disk(sb);
+		BUG_ON(ret);
+	}
+	return 0;
+}
+
 int write_ctree_super(struct btrfs_trans_handle *trans,
 		      struct btrfs_root *root)
 {
 	int ret;
 	struct btrfs_root *tree_root = root->fs_info->tree_root;
+	struct btrfs_root *chunk_root = root->fs_info->chunk_root;
+
+	if (root->fs_info->readonly)
+		return 0;
+
 	btrfs_set_super_generation(&root->fs_info->super_copy,
 				   trans->transid);
 	btrfs_set_super_root(&root->fs_info->super_copy,
 			     tree_root->node->start);
 	btrfs_set_super_root_level(&root->fs_info->super_copy,
 				   btrfs_header_level(tree_root->node));
+	btrfs_set_super_chunk_root(&root->fs_info->super_copy,
+				   chunk_root->node->start);
+	btrfs_set_super_chunk_root_level(&root->fs_info->super_copy,
+					 btrfs_header_level(chunk_root->node));
 	write_extent_buffer(root->fs_info->sb_buffer,
 			    &root->fs_info->super_copy, 0,
 			    sizeof(root->fs_info->super_copy));
-	ret = write_tree_block(trans, root, root->fs_info->sb_buffer);
+	ret = write_all_supers(root);
 	if (ret)
 		fprintf(stderr, "failed to write new super block err %d\n", ret);
 	return ret;
 }
 
+static int close_all_devices(struct btrfs_fs_info *fs_info)
+{
+	struct list_head *list;
+	struct list_head *next;
+	struct btrfs_device *device;
+
+	return 0;
+
+	list = &fs_info->fs_devices->devices;
+	list_for_each(next, list) {
+		device = list_entry(next, struct btrfs_device, dev_list);
+		close(device->fd);
+	}
+	return 0;
+}
+
 int close_ctree(struct btrfs_root *root)
 {
 	int ret;
@@ -487,7 +747,6 @@ int close_ctree(struct btrfs_root *root)
 	write_ctree_super(trans, root);
 	btrfs_free_transaction(root, trans);
 	btrfs_free_block_groups(root->fs_info);
-	close(root->fs_info->fp);
 	if (root->node)
 		free_extent_buffer(root->node);
 	if (root->fs_info->extent_root->node)
@@ -497,16 +756,25 @@ int close_ctree(struct btrfs_root *root)
 	free_extent_buffer(root->commit_root);
 	free_extent_buffer(root->fs_info->sb_buffer);
 
-	extent_map_tree_cleanup(&fs_info->extent_cache);
-	extent_map_tree_cleanup(&fs_info->free_space_cache);
-	extent_map_tree_cleanup(&fs_info->pending_tree);
-	extent_map_tree_cleanup(&fs_info->pinned_extents);
-	extent_map_tree_cleanup(&fs_info->del_pending);
-	extent_map_tree_cleanup(&fs_info->block_group_cache);
+	if (root->fs_info->chunk_root->node);
+		free_extent_buffer(root->fs_info->chunk_root->node);
+
+	if (root->fs_info->dev_root->node);
+		free_extent_buffer(root->fs_info->dev_root->node);
+
+	close_all_devices(root->fs_info);
+	extent_io_tree_cleanup(&fs_info->extent_cache);
+	extent_io_tree_cleanup(&fs_info->free_space_cache);
+	extent_io_tree_cleanup(&fs_info->block_group_cache);
+	extent_io_tree_cleanup(&fs_info->pinned_extents);
+	extent_io_tree_cleanup(&fs_info->pending_del);
+	extent_io_tree_cleanup(&fs_info->extent_ins);
 
 	free(fs_info->tree_root);
 	free(fs_info->extent_root);
 	free(fs_info->fs_root);
+	free(fs_info->chunk_root);
+	free(fs_info->dev_root);
 	free(fs_info);
 
 	return 0;
@@ -529,9 +797,16 @@ void btrfs_mark_buffer_dirty(struct extent_buffer *eb)
 	set_extent_buffer_dirty(eb);
 }
 
-int btrfs_buffer_uptodate(struct extent_buffer *eb)
+int btrfs_buffer_uptodate(struct extent_buffer *buf, u64 parent_transid)
 {
-	return extent_buffer_uptodate(eb);
+	int ret;
+
+	ret = extent_buffer_uptodate(buf);
+	if (!ret)
+		return ret;
+
+	ret = verify_parent_transid(buf->tree, buf, parent_transid);
+	return !ret;
 }
 
 int btrfs_set_buffer_uptodate(struct extent_buffer *eb)
