@@ -18,6 +18,7 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <uuid/uuid.h>
 #include "kerncompat.h"
 #include "radix-tree.h"
 #include "ctree.h"
@@ -60,14 +61,16 @@ static int print_inode_ref_item(struct extent_buffer *eb, struct btrfs_item *ite
 	u32 cur = 0;
 	u32 len;
 	u32 name_len;
+	u64 index;
 	char namebuf[BTRFS_NAME_LEN];
 	total = btrfs_item_size(eb, item);
 	while(cur < total) {
 		name_len = btrfs_inode_ref_name_len(eb, ref);
+		index = btrfs_inode_ref_index(eb, ref);
 		len = (name_len <= sizeof(namebuf))? name_len: sizeof(namebuf);
 		read_extent_buffer(eb, namebuf, (unsigned long)(ref + 1), len);
-		printf("\t\tinode ref namelen %u name: %.*s\n",
-		       name_len, len, namebuf);
+		printf("\t\tinode ref index %llu namelen %u name: %.*s\n",
+		       index, name_len, len, namebuf);
 		len = sizeof(*ref) + name_len;
 		ref = (struct btrfs_inode_ref *)((char *)ref + len);
 		cur += len;
@@ -75,6 +78,52 @@ static int print_inode_ref_item(struct extent_buffer *eb, struct btrfs_item *ite
 	return 0;
 }
 
+static void print_chunk(struct extent_buffer *eb, struct btrfs_chunk *chunk)
+{
+	int num_stripes = btrfs_chunk_num_stripes(eb, chunk);
+	int i;
+	printf("\t\tchunk length %llu owner %llu type %llu num_stripes %d\n",
+	       (unsigned long long)btrfs_chunk_length(eb, chunk),
+	       (unsigned long long)btrfs_chunk_owner(eb, chunk),
+	       (unsigned long long)btrfs_chunk_type(eb, chunk),
+	       num_stripes);
+	for (i = 0 ; i < num_stripes ; i++) {
+		printf("\t\t\tstripe %d devid %llu offset %llu\n", i,
+		      (unsigned long long)btrfs_stripe_devid_nr(eb, chunk, i),
+		      (unsigned long long)btrfs_stripe_offset_nr(eb, chunk, i));
+	}
+}
+static void print_dev_item(struct extent_buffer *eb,
+			   struct btrfs_dev_item *dev_item)
+{
+	printf("\t\tdev item devid %llu "
+	       "total_bytes %llu bytes used %Lu\n",
+	       (unsigned long long)btrfs_device_id(eb, dev_item),
+	       (unsigned long long)btrfs_device_total_bytes(eb, dev_item),
+	       (unsigned long long)btrfs_device_bytes_used(eb, dev_item));
+}
+
+static void print_uuids(struct extent_buffer *eb)
+{
+	char fs_uuid[37];
+	char chunk_uuid[37];
+	u8 disk_uuid[BTRFS_UUID_SIZE];
+
+	read_extent_buffer(eb, disk_uuid, (unsigned long)btrfs_header_fsid(eb),
+			   BTRFS_FSID_SIZE);
+
+	fs_uuid[36] = '\0';
+	uuid_unparse(disk_uuid, fs_uuid);
+
+	read_extent_buffer(eb, disk_uuid,
+			   (unsigned long)btrfs_header_chunk_tree_uuid(eb),
+			   BTRFS_UUID_SIZE);
+
+	chunk_uuid[36] = '\0';
+	uuid_unparse(disk_uuid, chunk_uuid);
+	printf("fs uuid %s\nchunk uuid %s\n", fs_uuid, chunk_uuid);
+}
+
 void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *l)
 {
 	int i;
@@ -89,6 +138,7 @@ void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *l)
 	struct btrfs_block_group_item *bi;
 	struct btrfs_extent_ref *ref;
 	struct btrfs_inode_ref *iref;
+	struct btrfs_dev_extent *dev_extent;
 	struct btrfs_disk_key disk_key;
 	struct btrfs_root_item root_item;
 	struct btrfs_block_group_item bg_item;
@@ -100,6 +150,7 @@ void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *l)
 		btrfs_leaf_free_space(root, l),
 		(unsigned long long)btrfs_header_generation(l),
 		(unsigned long long)btrfs_header_owner(l));
+	print_uuids(l);
 	fflush(stdout);
 	for (i = 0 ; i < nr ; i++) {
 		item = btrfs_item_nr(l, i);
@@ -132,6 +183,9 @@ void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *l)
 			di = btrfs_item_ptr(l, i, struct btrfs_dir_item);
 			print_dir_item(l, item, di);
 			break;
+		case BTRFS_ORPHAN_ITEM_KEY:
+			printf("\t\torphan item\n");
+			break;
 		case BTRFS_ROOT_ITEM_KEY:
 			ri = btrfs_item_ptr(l, i, struct btrfs_root_item);
 			read_extent_buffer(l, &root_item, (unsigned long)ri, sizeof(root_item));
@@ -190,9 +244,31 @@ void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *l)
 					    struct btrfs_block_group_item);
 			read_extent_buffer(l, &bg_item, (unsigned long)bi,
 					   sizeof(bg_item));
-			printf("\t\tblock group used %llu flags %x\n",
+			printf("\t\tblock group used %llu flags %llu\n",
 			       (unsigned long long)btrfs_block_group_used(&bg_item),
-			       bg_item.flags);
+			       (unsigned long long)btrfs_block_group_flags(&bg_item));
+			break;
+		case BTRFS_CHUNK_ITEM_KEY:
+			print_chunk(l, btrfs_item_ptr(l, i, struct btrfs_chunk));
+			break;
+		case BTRFS_DEV_ITEM_KEY:
+			print_dev_item(l, btrfs_item_ptr(l, i,
+					struct btrfs_dev_item));
+			break;
+		case BTRFS_DEV_EXTENT_KEY:
+			dev_extent = btrfs_item_ptr(l, i,
+						    struct btrfs_dev_extent);
+			printf("\t\tdev extent chunk_tree %llu\n"
+			       "\t\tchunk objectid %llu chunk offset %llu "
+			       "length %llu\n",
+			       (unsigned long long)
+			       btrfs_dev_extent_chunk_tree(l, dev_extent),
+			       (unsigned long long)
+			       btrfs_dev_extent_chunk_objectid(l, dev_extent),
+			       (unsigned long long)
+			       btrfs_dev_extent_chunk_offset(l, dev_extent),
+			       (unsigned long long)
+			       btrfs_dev_extent_length(l, dev_extent));
 			break;
 		case BTRFS_STRING_ITEM_KEY:
 			/* dirty, but it's simple */
@@ -203,6 +279,7 @@ void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *l)
 		fflush(stdout);
 	}
 }
+
 void btrfs_print_tree(struct btrfs_root *root, struct extent_buffer *eb)
 {
 	int i;
@@ -223,6 +300,7 @@ void btrfs_print_tree(struct btrfs_root *root, struct extent_buffer *eb)
 		(u32)BTRFS_NODEPTRS_PER_BLOCK(root) - nr,
 		(unsigned long long)btrfs_header_generation(eb),
 		(unsigned long long)btrfs_header_owner(eb));
+	print_uuids(eb);
 	fflush(stdout);
 	size = btrfs_level_size(root, btrfs_header_level(eb) - 1);
 	for (i = 0; i < nr; i++) {
@@ -241,7 +319,8 @@ void btrfs_print_tree(struct btrfs_root *root, struct extent_buffer *eb)
 	for (i = 0; i < nr; i++) {
 		struct extent_buffer *next = read_tree_block(root,
 					     btrfs_node_blockptr(eb, i),
-					     size);
+					     size,
+					     btrfs_node_ptr_generation(eb, i));
 		if (btrfs_is_leaf(next) &&
 		    btrfs_header_level(eb) != 1)
 			BUG();
